/* dim_op.c : engine of operations along a dimension (e.g. running mean)
 */

#include<stdio.h>
#include<string.h>
#include "ruby.h"
#include "narray.h"

#ifndef RARRAY_PTR
#  define RARRAY_PTR(ary) (RARRAY(ary)->ptr)
#endif
#ifndef RARRAY_LEN
#  define RARRAY_LEN(ary) (RARRAY(ary)->len)
#endif

#ifndef HAVE_NA_SHAPE_T
typedef int na_shape_t;
#endif

enum bc_type {
    BC_SIMPLE=10, 
    BC_CYCLIC=11, 
    BC_TRIM=12
};

#define ID3(i,j,k) ((i) + (j)*n0 + (k)*n0*n1)
#define ID3T(i,j,k) ((i) + (j)*n0 + (k)*n0*(n1-wlen+1))
#define ID3o(i,j,k) ((i) + (j)*n0 + (k)*n0*n1o)

static int
convol_result_length(na_shape_t n1, na_shape_t wlen, int ibc)
{
    int n1o;
    switch (ibc) {
    case BC_SIMPLE:
    case BC_CYCLIC:
	n1o = n1;
	break;
    case BC_TRIM:
	n1o = n1-wlen+1;
	break;
    }
    return n1o;
}

static void 
running_mean_nomiss(zi,n0,n1,n2, w,wlen, ibc, zo)
    // IN
    double *zi;
    na_shape_t n0;
    na_shape_t n1;
    na_shape_t n2;
    double *w;
    na_shape_t wlen;
    int ibc;
    // OUT
    double *zo;   //  must have been set to 0 everywhere
{
    na_shape_t i,j,k,m, jj;
    na_shape_t wl2s, wl2;
    double wsum;
    double *wc;     // scaled w (sum wc == 1)
    double *wcsum1,*wcsum2;  // for BC_SIMPLE; for shorter summation at edges
    double wcm;     // modified wcm for shorter summation

    wl2s = (wlen - 1)/2;  // e.g. 7->3, 6->2
    wl2 = wlen/2;         // e.g. 7->3, 6->3;   wl2s + wl2 == wlen - 1

    // < scale w to make its sum == 1 >

    for (m=0,wsum=0.0; m<wlen; m++) {
	wsum += w[m];
    }
    if (wsum != 1.0) {
	wc = ALLOCA_N(double, wlen);
	for (m=0; m<wlen; m++) {
	    wc[m] = w[m] / wsum;
	}
    } else {
	wc = w;
    }

    // < weighted running mean >

    switch (ibc) {
    case BC_SIMPLE:
	wcsum1 = ALLOCA_N(double, wl2s);
	wcsum1[0] = 1.0 - wc[0];              // sum of wc[1...wlen]
	for (m=1; m<wl2s; m++) {
	    wcsum1[m] = wcsum1[m-1] - wc[m];   // sum of wc[m+1...wlen]
	}
	wcsum2 = ALLOCA_N(double, wl2);
	wcsum2[0] = 1.0 - wc[wlen-1];         // sum of wc[0...wlen-1]
	for (m=wlen-2; m>wlen-1-wl2; m--) {
	    wcsum2[wlen-1-m] = wcsum2[wlen-2-m] - wc[m];   // sum of wc[0...m]
	}       // 1..wl2-1 with m

	for (k=0; k<n2; k++) {
	    // where j is close to the "left" end
	    for (j=0; j<wl2s; j++) {
		for (m=-j+wl2s; m<wlen; m++) {
		    wcm = wc[m] / wcsum1[wl2s-1-j];
		    for (i=0; i<n0; i++) {
			zo[ID3(i,j,k)] += wcm*zi[ID3(i,j-wl2s+m,k)];
		    }                                //^^^^^^^^ from 0
		}
	    }
	    // where grids are enough on both sides
	    for (j=wl2s; j<n1-wl2; j++) {
		for (m=0; m<wlen; m++) {
		    for (i=0; i<n0; i++) {
			zo[ID3(i,j,k)] += wc[m]*zi[ID3(i,j-wl2s+m,k)];
		    }
		}
	    }
	    // where j is close to the "right" end
	    for (j=n1-wl2; j<n1; j++) {
		for (m=0; m<n1-j+wl2s; m++) {
		    wcm = wc[m] / wcsum2[j-n1+wl2];  // wcsum2: 0...wl2
		    for (i=0; i<n0; i++) {
			zo[ID3(i,j,k)] += wcm*zi[ID3(i,j-wl2s+m,k)];
		    }                                //^^^^^^^^ upto n1-1
		}
	    }
	}
	break;
    case BC_CYCLIC:
	for (k=0; k<n2; k++) {
	    for (j=0; j<n1; j++) {
		for (m=0; m<wlen; m++) {
		    jj = j-wl2s+m;
		    if ( jj < 0 ) {
			jj += n1;
		    } else {
			jj %= n1;
		    }
		    for (i=0; i<n0; i++) {
			zo[ID3(i,j,k)] += wc[m]*zi[ID3(i,jj,k)];
		    }
		}
	    }
	}
	break;
    case BC_TRIM:
	// trim where extra grids are not enough
	for (k=0; k<n2; k++) {
	    for (j=0; j<n1-wlen+1; j++) {
		for (m=0; m<wlen; m++) {
		    for (i=0; i<n0; i++) {
			zo[ID3T(i,j,k)] += wc[m]*zi[ID3(i,j+m,k)];
		    }
		}
	    }
	}
	break;
    default:
	rb_raise(rb_eArgError,"Undefined boundary condision (%d)", ibc);
	break;
    }
}


static void 
running_mean_miss(zi,n0,n1,n2, w,wlen, ibc, zmiss,nminvalid, zo)
    // IN
    double *zi;
    na_shape_t n0;
    na_shape_t n1;
    na_shape_t n2;
    double *w;
    na_shape_t wlen;
    int ibc;
    double zmiss;
    int nminvalid;
    // OUT
    double *zo;   //  must have been set to 0 everywhere
{
    na_shape_t i,j,k,m, jj;
    na_shape_t wl2s, wl2;
    double wsum;
    int nm;

    wl2s = (wlen - 1)/2;  // e.g. 7->3, 6->2
    wl2 = wlen/2;         // e.g. 7->3, 6->3;   wl2s + wl2 == wlen - 1

    // < weighted running mean >

    switch (ibc) {
    case BC_SIMPLE:
	for (k=0; k<n2; k++) {
	    // where j is close to the "left" end
	    for (j=0; j<wl2s; j++) {
		for (i=0; i<n0; i++) {
		    wsum = 0.0;
		    nm = 0;
		    for (m=-j+wl2s; m<wlen; m++) {
			if ( zi[ID3(i,j-wl2s+m,k)] != zmiss ) {
			    zo[ID3(i,j,k)] += w[m]*zi[ID3(i,j-wl2s+m,k)];
                                                          //^^^^^^^^ from 0
			    wsum += w[m];
			    nm++;
			}
		    }
		    if (nm >= nminvalid) {
			zo[ID3(i,j,k)] /= wsum;
		    } else {
			zo[ID3(i,j,k)] = zmiss;
		    }
		}
	    }
	    // where grids are enough on both sides
	    for (j=wl2s; j<n1-wl2; j++) {
		for (i=0; i<n0; i++) {
		    wsum = 0.0;
		    nm = 0;
		    for (m=0; m<wlen; m++) {
			if ( zi[ID3(i,j-wl2s+m,k)] != zmiss ) {
			    zo[ID3(i,j,k)] += w[m]*zi[ID3(i,j-wl2s+m,k)];
			    wsum += w[m];
			    nm++;
			}
		    }
		    if (nm >= nminvalid) {
			zo[ID3(i,j,k)] /= wsum;
		    } else {
			zo[ID3(i,j,k)] = zmiss;
		    }
		}
	    }
	    // where j is close to the "right" end
	    for (j=n1-wl2; j<n1; j++) {
		for (i=0; i<n0; i++) {
		    wsum = 0.0;
		    nm = 0;
		    for (m=0; m<n1-j+wl2s; m++) {
			if ( zi[ID3(i,j-wl2s+m,k)] != zmiss ) {
			    zo[ID3(i,j,k)] += w[m]*zi[ID3(i,j-wl2s+m,k)];
		                                         //^^^^^^^^ upto n1-1
			    wsum += w[m];
			    nm++;
			}
		    }
		    if (nm >= nminvalid) {
			zo[ID3(i,j,k)] /= wsum;
		    } else {
			zo[ID3(i,j,k)] = zmiss;
		    }
		}
	    }
	}
	break;
    case BC_CYCLIC:
	for (k=0; k<n2; k++) {
	    for (j=0; j<n1; j++) {
		for (i=0; i<n0; i++) {
		    wsum = 0.0;
		    nm = 0;
		    for (m=0; m<wlen; m++) {
			jj = j-wl2s+m;
			if ( jj < 0 ) {
			    jj += n1;
			} else {
			    jj %= n1;
			}
			if ( zi[ID3(i,jj,k)] != zmiss ) {
			    zo[ID3(i,j,k)] += w[m]*zi[ID3(i,jj,k)];
			    wsum += w[m];
			    nm++;
			}
		    }
		    if (nm >= nminvalid) {
			zo[ID3(i,j,k)] /= wsum;
		    } else {
			zo[ID3(i,j,k)] = zmiss;
		    }
		}
	    }
	}
	break;
    case BC_TRIM:
	// trim where extra grids are not enough
	for (k=0; k<n2; k++) {
	    for (j=0; j<n1-wlen+1; j++) {
		for (i=0; i<n0; i++) {
		    wsum = 0.0;
		    nm = 0;
		    for (m=0; m<wlen; m++) {
			if ( zi[ID3(i,j+m,k)] != zmiss ) {
			    zo[ID3T(i,j,k)] += w[m]*zi[ID3(i,j+m,k)];
			    wsum += w[m];
			    nm++;
			}
		    }
		    if (nm >= nminvalid) {
			zo[ID3T(i,j,k)] /= wsum;
		    } else {
			zo[ID3T(i,j,k)] = zmiss;
		    }
		}
	    }
	}
	break;
    default:
	rb_raise(rb_eArgError,"Undefined boundary condision (%d)", ibc);
	break;
    }
}

// 注：convolution も引数は同じなのでドライバーは兼ねられる
// (入り口はわけないとならない)が今のところやってない．
// 
static VALUE
running_mean(int argc, VALUE *argv, VALUE self)
{
    VALUE vi;    // mandatory 1st arg; input NArray
    VALUE dim;   // mandatory 2nd arg; Integer 
    VALUE wgt;   // mandatory 3rd arg; weight 1D NArray
    VALUE bc;    // mandatory 4th arg; boundary condition (Integer class const)
    VALUE missv=Qnil; // optional 5th arg; if present(Float) vi may have missing
    int nminvalid=1;     // optional 6th arg; miminum count of non-missing vales
    int with_miss;
    struct NARRAY *na;
    na_shape_t *shi, wlen;
    na_shape_t n0, n1, n2;
    double *zi, *w, zmiss;
    int d, i, rank, ibc;
    VALUE vo;
    na_shape_t *sho, n1o;
    double *zo;

    // < process arguments >

    if ( argc<4 || argc>6 ) { rb_raise(rb_eArgError,"Need 4 to 6 arguments"); }
    vi  = argv[0];
    dim = argv[1];
    wgt = argv[2];
    bc  = argv[3];
    with_miss = (argc > 4) && (!NIL_P(argv[4]));
    if ( with_miss ) { missv = argv[4]; }
    if ( argc == 6) {
	nminvalid = NUM2INT( argv[5] );
    }

    //   1st arg
    if (!IsNArray(vi)) { rb_raise(rb_eArgError,"1st arg must be a NArray"); }
    vi = na_cast_object(vi, NA_DFLOAT);
    rank = NA_RANK(vi);
    zi = NA_PTR_TYPE(vi, double *);
    GetNArray(vi, na);
    shi = na->shape;

    //   2nd arg
    d = NUM2INT( dim );

    //   3rd arg
    if (!IsNArray(wgt)) {rb_raise(rb_eArgError,"3rd arg must be a 1D NArray");}
    if (NA_RANK(wgt)!=1) {rb_raise(rb_eArgError,"3rd arg must be a 1D NArray");}
    wgt = na_cast_object(wgt, NA_DFLOAT);
    wlen = NA_TOTAL(wgt);
    w = NA_PTR_TYPE(wgt, double *);

    //   4th arg
    ibc = NUM2INT( bc );

    //   5th arg
    if ( with_miss ) { zmiss = NUM2DBL(missv); }

    //   6th arg
    if (nminvalid > wlen) {rb_raise(rb_eArgError,"nminvalid > filtering length");}

    // < shape as 3D >

    n1 = shi[d];   // length of the dim to filter
    if (wlen >= n1) {rb_raise(rb_eArgError,"filter len >= len of the dim");}

    n0 = n2 = 1;
    for (i=0; i<d; i++) {
	n0 *= shi[i];   // total lengths of dims before d
    }
    for (i=d+1; i<rank; i++) {
	n2 *= shi[i];   // total lengths of dims after d
    }

    // < initialize the NArray to ruturn >

    n1o = convol_result_length(n1, wlen, ibc);

    sho = ALLOCA_N(int, rank);
    for(i=0; i<rank; i++){
	sho[i] = ( i!=d ? shi[i] : n1o );
    }

    vo = na_make_object(NA_DFLOAT, rank, sho, cNArray);
    GetNArray(vo, na);
    na_clear_data(na);
    zo = NA_PTR_TYPE(vo, double *);

    // < do the job >
    if ( with_miss ) {
	running_mean_miss(zi,n0,n1,n2, w,wlen, ibc, zmiss, nminvalid, zo);
    } else {
	running_mean_nomiss(zi,n0,n1,n2, w,wlen, ibc, zo);
    }

    return vo;
}

static void 
bin_mean_nomiss(zi,n0,n1,n2, len, zo)
    // IN
    double *zi;
    na_shape_t n0;
    na_shape_t n1;
    na_shape_t n2;
    na_shape_t len;
    // OUT
    double *zo;   //  must have been set to 0 everywhere
{
    na_shape_t n1o;
    na_shape_t i,j,k,m;
    double fact;

    n1o = n1 / len;
    fact = 1.0 / len;

    for (k=0; k<n2; k++) {
	for (j=0; j<n1o; j++) {
	    for (m=0; m<len; m++) {
		for (i=0; i<n0; i++) {
		    zo[ID3o(i,j,k)] += fact*zi[ID3(i,j*len+m,k)];
		}
	    }
	}
    }
}

static void 
bin_sum_nomiss(zi,n0,n1,n2, len, zo)
    // IN
    double *zi;
    na_shape_t n0;
    na_shape_t n1;
    na_shape_t n2;
    na_shape_t len;
    // OUT
    double *zo;   //  must have been set to 0 everywhere
{
    na_shape_t n1o;
    na_shape_t i,j,k,m;

    n1o = n1 / len;

    for (k=0; k<n2; k++) {
	for (j=0; j<n1o; j++) {
	    for (m=0; m<len; m++) {
		for (i=0; i<n0; i++) {
		    zo[ID3o(i,j,k)] += zi[ID3(i,j*len+m,k)];
		}
	    }
	}
    }
}

static void 
bin_mean_miss(zi,n0,n1,n2, len, zmiss,nminvalid, zo)
    // IN
    double *zi;
    na_shape_t n0;
    na_shape_t n1;
    na_shape_t n2;
    na_shape_t len;
    double zmiss;
    int nminvalid;
    // OUT
    double *zo;   //  must have been set to 0 everywhere
{
    na_shape_t n1o;
    na_shape_t i,j,k,m;
    na_shape_t cnt;

    n1o = n1 / len;

    for (k=0; k<n2; k++) {
	for (j=0; j<n1o; j++) {
	    for (i=0; i<n0; i++) {
		cnt = 0;
		for (m=0; m<len; m++) {
		    if ( zi[ID3(i,j*len+m,k)] != zmiss ) {
			zo[ID3o(i,j,k)] += zi[ID3(i,j*len+m,k)];
			cnt++;
		    }
		}
		if (cnt >= nminvalid) {
		    zo[ID3o(i,j,k)] /= cnt;
		} else {
		    zo[ID3o(i,j,k)] = zmiss;
		}
	    }
	}
    }
}

static void 
bin_sum_miss(zi,n0,n1,n2, len, zmiss,nminvalid, zo)
    // IN
    double *zi;
    na_shape_t n0;
    na_shape_t n1;
    na_shape_t n2;
    na_shape_t len;
    double zmiss;
    int nminvalid;
    // OUT
    double *zo;   //  must have been set to 0 everywhere
{
    na_shape_t n1o;
    na_shape_t i,j,k,m;
    na_shape_t cnt;

    n1o = n1 / len;

    for (k=0; k<n2; k++) {
	for (j=0; j<n1o; j++) {
	    for (i=0; i<n0; i++) {
		cnt = 0;
		for (m=0; m<len; m++) {
		    if ( zi[ID3(i,j*len+m,k)] != zmiss ) {
			zo[ID3o(i,j,k)] += zi[ID3(i,j*len+m,k)];
			cnt++;
		    }
		}
		if ( !(cnt >= nminvalid) ) {
		    zo[ID3o(i,j,k)] = zmiss;
		}
	    }
	}
    }
}

// bin_mean or bin_sum depeiding on the last arg
static VALUE
bin_mean_sum(int argc, VALUE *argv, VALUE self, int mean)
{
    VALUE vi;   // 1st arg; input NArray
    VALUE dim;  // 2nd arg
    na_shape_t len;   // 3rd arg
    VALUE missv=Qnil; // optional 4th arg; if present(Float) vi may have missing
    int nminvalid=1;  // optional 5th arg

    int with_miss;
    struct NARRAY *na;
    double *zi, zmiss;
    na_shape_t *shi;
    na_shape_t n0, n1, n2;
    int d, i, rank;
    VALUE vo;
    na_shape_t *sho, n1o;
    double *zo;

    // < process arguments >

    if ( argc<3 || argc>5 ) { rb_raise(rb_eArgError,"Need 2 or 3 arguments"); }
    vi  = argv[0];

    dim = argv[1];
    d = NUM2INT( dim );

    len  = NUM2INT( argv[2] );
    if (len<1) {rb_raise(rb_eArgError,"len must be >= 1");}
    
    with_miss = (argc > 3) && (!NIL_P(argv[3]));
    if ( with_miss ) { 
	missv = argv[3];
	zmiss = NUM2DBL(missv);
    }

    if (argc==5) { 
	nminvalid  = NUM2INT( argv[4] );
	if (nminvalid > len) {rb_raise(rb_eArgError,"nminvalid > bin length");}
    }


    // < NArray elements >

    rank = NA_RANK(vi);
    zi = NA_PTR_TYPE(vi, double *);
    GetNArray(vi, na);
    shi = na->shape;

    // < shape as 3D >

    n1 = shi[d];   // length of the dim
    if (len >= n1) {rb_raise(rb_eArgError,"filter len >= len of the dim");}

    n0 = n2 = 1;
    for (i=0; i<d; i++) {
	n0 *= shi[i];   // total lengths of dims before d
    }
    for (i=d+1; i<rank; i++) {
	n2 *= shi[i];   // total lengths of dims after d
    }

    // < initialize the NArray to ruturn >

    n1o = n1 / len;

    sho = ALLOCA_N(int, rank);
    for(i=0; i<rank; i++){
	sho[i] = ( i!=d ? shi[i] : n1o );
    }

    vo = na_make_object(NA_DFLOAT, rank, sho, cNArray);
    GetNArray(vo, na);
    na_clear_data(na);
    zo = NA_PTR_TYPE(vo, double *);

    // < do the job >
    if ( mean )  {
	if ( with_miss ) {
	    bin_mean_miss(zi,n0,n1,n2, len, zmiss,nminvalid, zo);
	} else {
	    bin_mean_nomiss(zi,n0,n1,n2, len, zo);
	}
    } else {
	if ( with_miss ) {
	    bin_sum_miss(zi,n0,n1,n2, len, zmiss,nminvalid, zo);
	} else {
	    bin_sum_nomiss(zi,n0,n1,n2, len, zo);
	}
    }

    return vo;
}

static VALUE
bin_mean(int argc, VALUE *argv, VALUE self)
{
    bin_mean_sum(argc, argv, self, 1);
}

static VALUE
bin_sum(int argc, VALUE *argv, VALUE self)
{
    bin_mean_sum(argc, argv, self, 0);
}

void
init_gphys_dim_op()
{
    static VALUE mNumRu;
    static VALUE cGPhys;
    static VALUE cVArray;

    // rb_require("narray");  // it does not work ??
    mNumRu = rb_define_module("NumRu");

    cGPhys = rb_define_class_under(mNumRu, "GPhys", rb_cObject);
    rb_define_private_method(cGPhys, "c_running_mean", running_mean, -1);

    cVArray = rb_define_class_under(mNumRu, "VArray", rb_cObject);
    rb_define_private_method(cVArray, "c_bin_mean", bin_mean, -1);
    rb_define_private_method(cVArray, "c_bin_sum", bin_sum, -1);
}
